## @namespace unogenerator.commons
from colorama import Fore, Style
from datetime import datetime, date
from gettext import translation
from logging import info, ERROR, WARNING, INFO, DEBUG, CRITICAL, basicConfig, error
from pkg_resources import resource_filename
from psutil import process_iter
from uno import createUnoStruct
from unogenerator.reusing.listdict_functions import listdict_min
from time import sleep

__version__ = '0.21.0'
__versiondatetime__=datetime(2022, 3, 9, 12, 8)
__versiondate__=__versiondatetime__.date()

try:
    t=translation('unogenerator',resource_filename("unogenerator","locale"))
    _=t.gettext
except:
    _=str

class ColorsNamed:
    Black=0x111111
    White=0xFFFFFF
    Blue=0x9999ff
    Red=0xFF9999
    Green=0xc0FFc0
    Orange=0xffdca8
    Yellow=0xffffc0
    GrayLight=0xd2ced4
    GrayDark=0xa29ea4
    GrayVeryDark=0x726e74

def datetime2uno( dt):
    r=createUnoStruct("com.sun.star.util.DateTime")
    r.Year=dt.year
    r.Month=dt.month
    r.Day=dt.day
    r.Hours=dt.hour
    r.Minutes=dt.minute
    r.Seconds=dt.second
    return r

def date2uno( dt):
    r=createUnoStruct("com.sun.star.util.Date")
    r.Year=dt.year
    r.Month=dt.month
    r.Day=dt.day
    return r
    

## Function used in argparse_epilog
## @return String
def argparse_epilog():
    return _("Developed by Mariano Muñoz 2020-{}").format(__versiondate__.year)


## Allows to operate with columns letter names
## @param letter String with the column name. For example A or AA...
## @param number Columns to move
## @return String With the name of the column after movement
def columnAdd(letter, number):
    letter_value=column2number(letter)+number
    return number2column(letter_value)


def rowAdd(letter,number):
    return str(int(letter)+number)

## Convierte un número  con el numero de columna al nombre de la columna de hoja de datos
##
## Number to Excel-style column name, e.g., 1 = A, 26 = Z, 27 = AA, 703 = AAA.
def number2column(n):
    name = ''
    while n > 0:
        n, r = divmod (n - 1, 26)
        name = chr(r + ord('A')) + name
    return name

## Convierte una columna de hoja de datos a un número
##
## Excel-style column name to number, e.g., A = 1, Z = 26, AA = 27, AAA = 703.
def column2number(name):
    n = 0
    for c in name:
        n = n * 26 + 1 + ord(c) - ord('A')
    return n

## Converts a column name to a index position (number of column -1)
def column2index(name):
    return column2number(name)-1

## Convierte el nombre de la fila de la hoja de datos a un índice, es decir el número de la fila -1
def row2index(number):
    return int(number)-1

## Covierte el nombre de la fila de la hoja de datos a un  numero entero que corresponde con el numero de la fila
def row2number(strnumber):
    return int(strnumber)

## Convierte el numero de la fila al nombre de la fila en la hoja de datos , que corresponde con un string del numero de la fila
def number2row(number):
    return str(number)
    
## Convierte el indice de la fila al numero cadena de la hoja de datos
def index2row(index):
    return str(index+1)
    
## Convierte el indice de la columna a la cadena de letras de la columna de la hoja de datos
def index2column(index):
    return number2column(index+1)

## Class that manage spreadsheet coordinates (letter + number)
class Coord:
    def __init__(self, strcoord):
        self.letter, self.number=self.__extract(strcoord)

    def __repr__(self):
        return f"Coord <{self}>"
        
    def __str__(self):
        return self.string()

    def __eq__(self, b):
        b=Coord.assertCoord(b)
        if self.letter==b.letter and self.number==b.number:
            return True
        return False
        
    def __extract(self, strcoord):
        if strcoord.find(":")!=-1:
            print("I can't manage range coord")
            return
        letter=""
        number=""
        for l in strcoord:
            if l.isdigit()==False:
                letter=letter+l
            else:
                number=number+l
        return (letter,number)

    ## Returns Coord string
    ## @return string For example "Z1"
    def string(self):
        return self.letter+self.number

    ## Add a number of rows to the Coord and return itself
    ## @param num Integer Can be positive and negative. When num is negative, if Coord.number is less than 1, returns 1
    def addRow(self, num=1):
        if self.numberIndex()+num<0:
            self.number="1"
        else:
            self.number=rowAdd(self.number, num)
        return self

    ## Add a number of columns/letters to the Coord and return itself
    ## @param num Integer. Can be positive and negative. When num is negative, if Coord.letter is less than A, returns A.
    def addColumn(self, num=1):
        if self.letterIndex()+num<0:
            self.letter="A"
        else:
            self.letter=columnAdd(self.letter, num)
        return self

    ## Add a number of rows to the Coord and return a copy of the object
    ## @param num Integer Can be positive and negative. When num is negative, if Coord.number is less than 1, returns 1
    def addRowCopy(self, num=1):
        r=Coord(self.string())
        r.addRow(num)
        return r

    ## Add a number of columns/letters to the Coord and return a copy of the objject
    ## @param num Integer. Can be positive and negative. When num is negative, if Coord.letter is less than A, returns A.
    def addColumnCopy(self, num=1):
        r=Coord(self.string())
        r.addColumn(num)
        return r

    ## Reterns the letter/column index (letterPosition()-1)
    ## @return int
    def letterIndex(self):
        return column2index(self.letter)

    ## Returns the letter/column position
    ## @return int 
    def letterPosition(self):
        return column2number(self.letter)

    ## Returns the number/row index (numberPosition()-1)
    def numberIndex(self):
        return row2index(self.number)

    ## Returns the number/row position
    def numberPosition(self):
        return row2number(self.number)

    @staticmethod
    def assertCoord(o):
        if o.__class__==Coord:
            return o
        elif o.__class__==str:
            return Coord(o)
        else:
            error("{} is not a coord".format(o))
            



## Class that manages spreadsheet Ranges for ods and xlsx
class Range:
    def __init__(self,strrange):
        self.start, self.end=self.__extract(strrange)

    def __repr__(self):
        return f"Range <{self}>"
        
    def __str__(self):
        return self.string()
        
    @classmethod
    def from_coords_indexes(cls, start_letter_index, start_number_index, end_letter_index, end_number_index):
        c_start=Coord_from_index(start_letter_index, start_number_index)
        c_end=Coord_from_index(end_letter_index, end_number_index)
        return cls(f"{c_start}:{c_end}")

    ##Return the outcome of the test b in a. Note the reversed operands.
    def __contains__(self, b):
        if (    b.letterIndex()>=self.start.letterIndex() and 
                b.letterIndex()<=self.end.letterIndex() and 
                b.numberIndex()>=self.start.numberIndex() and 
                b.numberIndex()<=self.end.numberIndex())==True:
            return True
        return False

    ## Returns a list of rows of all Coord objects in the range
    def coords(self):
        r=[]
        for row in range(self.numRows()):
            tmprow=[]
            for column in range(self.numColumns()):
                tmprow.append(self.start.addRowCopy(row).addColumnCopy(column))
            r.append(tmprow)
        return r

    ## Converts a string to a Range. Returns None if conversion can't be done
    def __extract(self,range):
        if range.find(":")==-1:
            print("I can't manage this range")
            return
        a=range.split(":")
        return (Coord(a[0]), Coord(a[1]))

    ## String of a range in spreadsheets
    def string(self):
        return "{}:{}".format(self.start.string(), self.end.string())

    ## Number of range of the range
    def numRows(self):
        return row2number(self.end.number)-row2number(self.start.number) +1

    ## Number of columns of the range
    def numColumns(self):
        return column2number(self.end.letter)-column2number(self.start.letter) +1

    ## Returns a Range object even o is a str or a Range
    @staticmethod
    def assertRange(o):
        if o.__class__==Range:
            return o
        elif o.__class__==str:
            return Range(o)


    ## Adds a row to the end Coord, so it adds a row to the range
    def addRowAfter(self, num=1):
        self.end=self.end.addRow(num)
        return self

    def addRowAfterCopy(self, num=1):
        r=Range(self.string())
        r.addRowAfter(num)
        return r

    ## Adds a column to the end Coord, so it adds a column to the range
    def addColumnAfter(self, num=1):
        self.end=self.end.addColumn(num)
        return self

    def addColumnAfterCopy(self, num=1):
        r=Range(self.string())
        r.addColumnAfter(num)
        return r

    ## Adds a row to the top of the start Coord, so it adds a row to the range. If start Coord number is 1 returns the same Coord
    def addRowBefore(self, num=1):
        self.start=self.start.addRow(-num)
        return self

    def addRowBeforeCopy(self, num=1):
        r=Range(self.string())
        r.addRowBefore(num)
        return r

    ## Adds a column to the end Coord, so it adds a column to the range. If start Coord letter is A, returns the same Coord
    def addColumnBefore(self, num=1):
        self.start=self.start.addColumn(-num)
        return self

    def addColumnBeforeCopy(self, num=1):
        r=Range(self.string())
        r.addColumnBefore(num)
        return r

    ## Returns a list of tuples(column_index, row_index)
    def indexes_list(self, plain=False):
        r=[]
        if plain is True:
            for letter_index in range(self.start.letterIndex(), self.end.letterIndex()+1):
                for number_index in range(self.start.numberIndex(), self.end.numberIndex()+1):
                    r.append((letter_index, number_index))
        else:
            for number_index in range(self.start.numberIndex(), self.end.numberIndex()+1):
                row=[]
                for letter_index in range(self.start.letterIndex(), self.end.letterIndex()+1 ):
                    row.append((letter_index, number_index))
                r.append(row)
        return r

    def coords_list(self, plain=False):
        r=[]
        if plain is True:
            for letter_index, number_index in self.indexes_list():
                r.append(Coord_from_index(letter_index, number_index))
        else:
            for row  in self.indexes_list():
                r2=[]
                for  letter_index , number_index  in row:
                    r2.append(Coord_from_index(letter_index, number_index))
                r.append(r2)
        return r

## Sets debug sustem, needs
## @param args It's the result of a argparse     args=parser.parse_args()        
def addDebugSystem(level):
    logFormat = "%(asctime)s.%(msecs)03d %(levelname)s %(message)s [%(module)s:%(lineno)d]"
    dateFormat='%F %I:%M:%S'

    if level=="DEBUG":#Show detailed information that can help with program diagnosis and troubleshooting. CODE MARKS
        basicConfig(level=DEBUG, format=logFormat, datefmt=dateFormat)
    elif level=="INFO":#Everything is running as expected without any problem. TIME BENCHMARCKS
        basicConfig(level=INFO, format=logFormat, datefmt=dateFormat)
    elif level=="WARNING":#The program continues running, but something unexpected happened, which may lead to some problem down the road. THINGS TO DO
        basicConfig(level=WARNING, format=logFormat, datefmt=dateFormat)
    elif level=="ERROR":#The program fails to perform a certain function due to a bug.  SOMETHING BAD LOGIC
        basicConfig(level=ERROR, format=logFormat, datefmt=dateFormat)
    elif level=="CRITICAL":#The program encounters a serious error and may stop running. ERRORS
        basicConfig(level=CRITICAL, format=logFormat, datefmt=dateFormat)
    info("Debug level set to {}".format(level))



## Creates a Coord object from spreadsheet letters
def Coord_from_letters(column, letter):
    return Coord(column+letter)

## Creates a Coord object from spreadsheet index coords
def Coord_from_index(column_index, row_index):
    return Coord(index2column(column_index)+index2row(row_index))
    
## Creates a Range object from itself stard and end coords
def Range_from_coords(start, end):
    start=Coord.assertCoord(start)
    end=Coord.assertCoord(end)
    return Range(f"{start}:{end}")

def generate_formula_total_string(key, coord_from, coord_to):
    if key == "#SUM":
        s="=SUM({}:{})".format(coord_from.string(), coord_to.string())
    elif key == "#AVG":
        s="=AVERAGE({}:{})".format(coord_from.string(), coord_to.string())
    elif key == "#MEDIAN":
        s="=MEDIAN({}:{})".format(coord_from.string(), coord_to.string())
    else:
        s=key
    return s

def guess_object_style(o):
    if o is None:
        return "Normal"
    elif o.__class__.__name__=="int":
        return "Integer"
    elif o.__class__.__name__=="str":
        return "Normal"
    elif o.__class__.__name__ in ["Currency", "Money" ]:
        return o.currency
    elif o.__class__.__name__=="Percentage":
        return "Percentage"
    elif o.__class__.__name__ in ("Decimal", "float"):
        return  "Float2"
    elif o.__class__.__name__=="datetime":
        return "Datetime"
    elif o.__class__.__name__=="date":
        return "Date"
    elif o.__class__.__name__=="timedelta":
        return "Normal"
    elif o.__class__.__name__=="time":
        return "Time"
    elif o.__class__.__name__=="bool":
        return "Bool"
    else:
        info("guess_object_style not guessed {}".format( o.__class__))
        return "Bold"
        
def datetime2localc1989(o):
    delta = o -  datetime(1899, 12, 30)
    return float(delta.days) + float(delta.seconds) / 86400
    
def date2localc1989(o):
    delta = o -  date(1899, 12, 30)
    return float(delta.days) 

def time2localc1989(o):
    seconds=o.hour*3600+o.minute*60+o.second
    return float(seconds) / 86400
    
## Used to change port when there are multiple sockets of libreoffice accepting
def next_port(last,  first_port,  instances):
    if last==first_port+instances -1:
        return first_port
    else:
        return last+1

def get_range_from_iterable_object( coord_start, o):
    coord_start=Coord.assertCoord(coord_start)
    r=Range(f"{coord_start.string()}:{coord_start.string()}")

    if len(o)==0:
        return r

    if o[0].__class__.__name__ in ("list","dict", "OrderedDict"): #Iterables con len
        len_rows=len(o)-1
        len_columns=len(o[0])-1
    else:
        len_rows=0
        len_columns=len(o)-1
    return Range(f"{coord_start.string()}:{coord_start.addRowCopy(len_rows).addColumnCopy(len_columns).string()}")

## @param attempts (Integer). Sometimes when server is busy this method fails to detect info. So I make several attempts with a time interval
## @return listdict with process info

def get_from_process_info(cpu_percentage=False, attempts=10):
    for attempt in range(attempts):
        try:
            r=[]
            for p in process_iter(['name','cmdline', 'pid']): 
                d={}
                if p.info['name']=='soffice.bin':
                    if  'file:///tmp/unogenerator'  in ' '.join(p.info['cmdline']):
                        d["port"]=int(p.info['cmdline'][1][-4:])
                        d["pid"]=p.pid
                        d["mem"]=p.memory_info().rss
                        d["cpu_number"]=p.cpu_num()
                        if cpu_percentage is True:
                            d["cpu_percentage"]=p.cpu_percent(interval=0.01)
                            d["object"]=p
                        r.append(d)
            if len(r)==0:
                raise
            return r
        except:
            print(_(f"I couldn't detect unogenerator process info ({attempt}/{attempts} attempts)"))
            sleep(5)
            continue
    print(_("Have you launched unogenerator instances?. Please run unogenerator_start"))
    return []
    

def get_from_process_numinstances_and_firstport():        
    ld=get_from_process_info()
    if len(ld)==0:
        print(_("I couldn't detect unogenerator instances"))
    return len(ld), listdict_min(ld,"port")
   
def is_formula(s):
    if s is None:
        return False
    s=str(s)
    if s.startswith("=") or s.startswith("+"):
        return True
    return False
   
def red(s):
    return Style.BRIGHT + Fore.RED + str(s) + Style.RESET_ALL
def green(s):
    return Style.BRIGHT + Fore.GREEN + str(s) + Style.RESET_ALL
def magenta(s):
    return Style.BRIGHT + Fore.MAGENTA + str(s) + Style.RESET_ALL



from unogenerator import ODS
from unogenerator.unogenerator import createUnoService
from com.sun.star.beans import PropertyValue
from com.sun.star.sheet.ConditionEntryType import COLORSCALE
from uno import createUnoStruct
doc=ODS("colorscale.ods")

myCells=doc.sheet.getCellRangeByName("A1:A8")
##        print(unorange, dir(unorange))
myConditionalFormats = doc.sheet.ConditionalFormats.getConditionalFormats()
mcf=myConditionalFormats[0]
print("CONDITIONAL FORMATS", myConditionalFormats, dir(myConditionalFormats))
print()
print("CONDITIONAL FORMAT [0]", mcf, dir(mcf))

print(mcf.Count)
for o in mcf:
    print("ENTRY", o, dir(o))
    psi=o.getPropertySetInfo()
    print("PSI", psi, dir(psi),"\n")
    print(o.getPropertySetInfo())

    for p in psi.Properties:
        print("P", p, dir(p), p.typeName, p.value ,"\n", dir(p.value), "\n")



    for cse in o.ColorScaleEntries:
        print("cse",cse, dir(cse))
        print(cse.getFormula(), cse.getColor(),cse.getType())
    print("GPV", o.getPropertyValue("ColorScaleEntries"), dir(o.getPropertyValue("ColorScaleEntries")), "\n")


# cse=mcf
# print("CSE", cse, dir(cse))

##print(myCells.ConditionalFormat, dir(myCells.ConditionalFormat),  myCells.ConditionalFormat.getTypes())
##print(myCells.ConditionalFormat.getSupportedServiceNames())
##for e in myCells.ConditionalFormat.createEnumeration():
##    print(e, dir(e))

#myNewCells
#doc.addColumnWithStyle("B1", (1, 2, 4, 7, 5, 3, 6, 1))
#myNewCells=doc.sheet.getCellRangeByName("B1:B8")
#print(myNewCells)

#print("SHEET",  doc.sheet,  dir(doc.sheet))
#print()
#cursor=doc.sheet.createCursorByRange(doc.sheet.getCellRangeByName("A1:A8"))
#print("CURSOR", cursor, dir(cursor))
#print()
##print("CF", cursor.ConditionalFormat,  dir(cursor.ConditionalFormat), cursor.ConditionalFormat.Count)
#print(cursor.getRangeAddress())

# oRangos = doc.document.createInstance("com.sun.star.sheet.SheetCellRanges")
# oRangos.addRangeAddress( doc.sheet.getCellRangeByName( "B1:B8" ).getRangeAddress() ,False )
# doc.document.getCurrentController().select(oRangos)

# print("oRangos", oRangos, dir(oRangos))
# orcf=oRangos.ConditionalFormat
# print("oRangosConditionalFormat", orcf, dir(orcf))

# arr=(
#     PropertyValue('FilterName',0,'writer8',0),
# )


# new=orcf.addNew()
# print("new", new, dir(new))



# cfs=doc.sheet.ConditionalFormats.createByRange(oRangos)
# #cfo=doc.sheet.ConditionalFormats.getConditionalFormats()[0]
# #print("CFO", cfo, dir(cfo), len(cfo))
# #
# #for e in range(cfo.Count):
# #    p=cfo.getByIndex(e)
# #    print("ORIGINAL", p,  dir(p))
# #    cse=p.ColorScaleEntries[0]
# #    print("COLORSCALEENTRIES",  cse ,  dir(cse))
# #    print("COLORSCALEENTRIESDETAILS", cse.getColor() ,  cse.getFormula(), cse.getType(), cse.getTypes())

# cf=doc.sheet.ConditionalFormats.getConditionalFormats()[cfs-1]
# print("CF", cf, dir(cf), len(cf))
#cf.createEntry(1, 0)#1 es colorscale

#entry=cf.getByIndex(0)
#print("ENTRY",  entry, dir(entry))
#
#new = createUnoStruct("com.sun.star.sheet.XColorScaleEntry")
#new.acquire()
##print("NEW", new, dir(new))
##newi=new.__init__()
##print(newi)
#
#
#print("Calorscaleentryies", entry.ColorScaleEntries,  dir(entry.ColorScaleEntries))
#entry.ColorScaleEntries=(new, )
#print("Calorscaleentryies", entry.ColorScaleEntries,  dir(entry.ColorScaleEntries))
#se_uno=createUnoStruct("com.sun.star.sheet.ColorScale")
#print(se_uno,  dir(se_uno), se_uno.__class__)
#print(se_uno.getFormula())ColorScaleEntries
#p=PropertyValue('FilterName',0,'writer8',0),
#se=entry.queryInterface()
#print("SE",  se,  dir(se), se.__class__)

# doc.export_pdf("salida.pdf")
# doc.save("salida.ods")
doc.close()



from argparse import ArgumentParser, RawTextHelpFormatter
from colorama import init as colorama_init
from gettext import translation
from humanize import naturalsize
from os import system, makedirs
from pkg_resources import resource_filename
from unogenerator.commons import __version__, argparse_epilog, addDebugSystem, get_from_process_info, green, red, magenta
from unogenerator.reusing.casts import list2string
from unogenerator.reusing.listdict_functions import listdict_sum, listdict_average, listdict2list, listdict_order_by
from unogenerator.reusing.percentage import Percentage
from subprocess import run
from time import sleep

try:
    t=translation('unogenerator', resource_filename("unogenerator","locale"))
    _=t.gettext
except:
    _=str

def server_start():#!/bin/bash
    parser=ArgumentParser(
        description=_('Launches libreoffice server to run unogenerator code'), 
        epilog=argparse_epilog(), 
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--instances', help="Number of concurrent libreoffice instances", action="store", type=int,  default=8)
    parser.add_argument('--first_port', help="First port assigned to server", action="store", type=int, default=2002)
    parser.add_argument('--backtrace', help="Backtrace logs", action="store_true", default=False)
    args=parser.parse_args()

    command_start(args.instances, args.first_port, args.backtrace)
    
def command_start(instances, first_port=2002, backtrace=False):
    print(_(f"Preparing {instances} libreoffice server instances from port {first_port}:"))
    if backtrace is True:
        makedirs("/var/log/unogenerator/", exist_ok=True)
        print(_("You can see instances output in /var/log/unogenerator"))
        backtrace="--backtrace"
    else:
        backtrace=""
    
    for i in range(instances):
        port=first_port+i
        if backtrace is True:
            command=f'loffice --accept="socket,host=localhost,port={port};urp;StarOffice.ServiceManager" -env:UserInstallation=file:///tmp/unogenerator{port} --headless {backtrace} > /var/log/unogenerator/unogenerator.{port}.log 2>&1 &'
        else:
            command=f'loffice --accept="socket,host=localhost,port={port};urp;StarOffice.ServiceManager" -env:UserInstallation=file:///tmp/unogenerator{port} --headless &'

        print(_(f"  - Launched instance in port {port}"))
        system(command)

def server_stop():
    parser=ArgumentParser(
        description=_('Stops libreoffice server to finish using unogenerator code'), 
        epilog=argparse_epilog(), 
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--debug', help="Debug program information", choices=["DEBUG","INFO","WARNING","ERROR","CRITICAL"], default="ERROR")

    args=parser.parse_args()

    addDebugSystem(args.debug)
    command_stop()
def command_stop():
    command="pkill -c -f ';urp;StarOffice.ServiceManager'"
    s=run(command ,shell=True, capture_output=True)
    print(_(f"  - Server was stopped killing {s.stdout.decode('UTF-8').strip()} processes"))
    
    
def monitor():    
    colorama_init()
    parser=ArgumentParser(
        description=_('Monitor unogenerator statistics'), 
        epilog=argparse_epilog(), 
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--debug', help="Debug program information", choices=["DEBUG","INFO","WARNING","ERROR","CRITICAL"], default="ERROR")
    parser.add_argument('--restart', help="Restart server when idle", action="store_true", default=False)
    parser.add_argument('--max_mem_multiplier', help="Restart server when idle", action="store", type=int, default=2)
    args=parser.parse_args()

    addDebugSystem(args.debug)
    command_monitor(args.restart, args.max_mem_multiplier)
    
def command_monitor(restart, max_mem_multiplier):
    
    ld=get_from_process_info(cpu_percentage=True)
    ld=listdict_order_by(ld, "port")
    instances=len(ld)
    max_mem_recommended=instances*440010752*max_mem_multiplier
    list_ports=listdict2list(ld, 'port', True)
    cpu_nums=listdict2list(ld, 'cpu_number', True)
    cpu_percentage=Percentage(listdict_average(ld, "cpu_percentage"), 100)
    mem_total=listdict_sum(ld, "mem")
    str_mem_total=green(naturalsize(mem_total)) if mem_total<max_mem_recommended else red(naturalsize(mem_total))
    str_cpu_percentage= green(cpu_percentage) if cpu_percentage.value==0 else red(cpu_percentage)

    print(_(f"Instances: {green(instances)}"))
    print(_(f"Ports used: {green(list2string(list_ports))}"))
    print(_(f"CPU used: {green(list2string(cpu_nums))}"))
    print(_(f"Memory used: {str_mem_total}"))
    print(_(f"Max memory recommended: {green(naturalsize(max_mem_recommended))}"))    
    print(_(f"CPU percentage: {str_cpu_percentage}"))
    
    total_cons=0
    for d in ld:
        for con in  d["object"].connections():
#            print(con.laddr,  con.raddr, con.status)
            if con.status=="ESTABLISHED":
                total_cons=total_cons+1
    
    str_connections= green(total_cons) if total_cons==0 else red(total_cons)
    print(_(f"Connections: {str_connections}"))
    if restart is True:
        if mem_total>max_mem_recommended:
            
            if total_cons==0 :
                print(magenta(_("Restarting server")))
                first_port=min(list_ports)
                command_stop()
                sleep(1)
                command_start(instances, first_port, False)
                sleep(10)
                print(magenta(_("Server restarted")))
                command_monitor(False, max_mem_multiplier)
            else:
                print(magenta(_("Server wasn't restarted because it's busy")))
                
        else:
            print(magenta(_("Server doesn't need to be restarted")))



## @namespace unogenerator.demo
## @brief Generate ODF example files
from uno import getComponentContext
getComponentContext()
import argparse
from collections import OrderedDict
from concurrent.futures import ProcessPoolExecutor, as_completed
from datetime import datetime, date, timedelta
from gettext import translation
from logging import info
from multiprocessing import cpu_count
from pkg_resources import resource_filename

from unogenerator.commons import __version__, addDebugSystem, argparse_epilog, ColorsNamed, Coord as C, next_port, get_from_process_numinstances_and_firstport
from unogenerator.reusing.currency import Currency
from unogenerator.reusing.percentage import Percentage
from unogenerator.unogenerator import ODT_Standard, ODS_Standard
from unogenerator.helpers import helper_title_values_total_row,helper_title_values_total_column, helper_totals_row, helper_totals_column, helper_totals_from_range, helper_list_of_ordereddicts, helper_list_of_dicts, helper_list_of_ordereddicts_with_totals, helper_ods_sheet_stylenames
from os import remove
from tqdm import tqdm

try:
    t=translation('unogenerator', resource_filename("unogenerator","locale"))
    _=t.gettext
except:
    _=str

def remove_without_errors(filename):
    try:
        remove(filename)
    except OSError as e:
        print(_("Error deleting: {} -> {}".format(filename, e.strerror)))

## If arguments is None, launches with sys.argc parameters. Entry point is toomanyfiles:main
## You can call with main(['--pretend']). It's equivalento to os.system('program --pretend')
## @param arguments is an array with parser arguments. For example: ['--argument','9']. 
def main(arguments=None):
    parser=argparse.ArgumentParser(prog='unogenerator', description=_('Create example files using unogenerator module'), epilog=argparse_epilog(), formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--debug', help=_("Debug program information"), choices=["DEBUG","INFO","WARNING","ERROR","CRITICAL"], default="ERROR")
    group= parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--create', help="Create demo files", action="store_true",default=False)
    group.add_argument('--remove', help="Remove demo files", action="store_true", default=False)
    args=parser.parse_args(arguments)

    addDebugSystem(args.debug)
        
    if args.remove==True:
            for language in ['es', 'en']:
                remove_without_errors(f"unogenerator_documentation_{language}.odt")
                remove_without_errors(f"unogenerator_documentation_{language}.docx")
                remove_without_errors(f"unogenerator_documentation_{language}.pdf")
                remove_without_errors(f"unogenerator_example_{language}.ods")
                remove_without_errors(f"unogenerator_example_{language}.xlsx")
                remove_without_errors(f"unogenerator_example_{language}.pdf")

    if args.create==True:
        start=datetime.now()
        futures=[]
        with ProcessPoolExecutor(max_workers=cpu_count()+1) as executor:
            for language in ['es', 'en']:
                futures.append(executor.submit(demo_ods_standard, language))
                futures.append(executor.submit(demo_odt_standard, language))

        for future in as_completed(futures):
            future.result()
        print(_("All process took {}".format(datetime.now()-start)))


def main_concurrent(arguments=None):
    parser=argparse.ArgumentParser(prog='unogenerator', description=_('Create example files using unogenerator module'), epilog=argparse_epilog(), formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--debug', help="Debug program information", choices=["DEBUG","INFO","WARNING","ERROR","CRITICAL"], default="ERROR")
    group= parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--create', help="Create demo files", action="store_true",default=False)
    group.add_argument('--remove', help="Remove demo files", action="store_true", default=False)
    parser.add_argument('--loops', help="Loops of documentation jobs", action="store", default=30,  type=int)
    args=parser.parse_args(arguments)

    num_instances, first_port=get_from_process_numinstances_and_firstport()
    addDebugSystem(args.debug)

    if args.remove==True:
            for i in range(args.loops):
                for language in ['es', 'en']:
                    remove_without_errors(f"unogenerator_documentation_{language}.{i}.odt")
                    remove_without_errors(f"unogenerator_documentation_{language}.{i}.docx")
                    remove_without_errors(f"unogenerator_documentation_{language}.{i}.pdf")
                    remove_without_errors(f"unogenerator_example_{language}.{i}.ods")
                    remove_without_errors(f"unogenerator_example_{language}.{i}.xlsx")
                    remove_without_errors(f"unogenerator_example_{language}.{i}.pdf")

    if args.create==True:
        print(_(f"Launching concurrent demo with {num_instances} workers to a daemon with {num_instances} instances from {first_port} port"))

        start=datetime.now()
        futures=[]
        port=first_port
        with ProcessPoolExecutor(max_workers=num_instances) as executor:
            with tqdm(total=args.loops*4) as progress:
                for i in range(args.loops):
                    for language in ['es', 'en']:
                        port=next_port(port, first_port, num_instances)
                        future=executor.submit(demo_ods_standard, language, port, f".{i}")
                        future.add_done_callback(lambda p: progress.update())
                        futures.append(future)
                        port=next_port(port, first_port, num_instances)
                        future=executor.submit(demo_odt_standard, language, port, f".{i}")
                        future.add_done_callback(lambda p: progress.update())
                        futures.append(future)

                for future in as_completed(futures):
                    future.result()
            


        results = []
        for future in futures:
            result = future.result()
            results.append(result)
        print(_("All process took {}".format(datetime.now()-start)))

       
def demo_ods_standard(language, port=2002, suffix="",):
    lang1=translation('unogenerator', resource_filename("unogenerator","locale"), languages=[language])
    lang1.install()
    _=lang1.gettext
    
    doc=ODS_Standard(port)
    doc.setMetadata(
        _("UnoGenerator ODS example"),  
        _("Demo with ODS class"), 
        "Turulomio", 
        _(f"This file have been generated with UnoGenerator-{__version__}. You can see UnoGenerator main page in http://github.com/turulomio/unogenerator"), 
        ["unogenerator", "demo", "files"]
    )
    doc.createSheet("Styles")
    doc.setColumnsWidth([3.5, 5, 2, 2, 2, 2, 2, 5, 5, 3, 3])
    
    
    doc.setCellName("A1",  "MYNAME")
    
    doc.addCellWithStyle("A1", _("Style name"), ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("B1", _("Date and time"), ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("C1", _("Date"), ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("D1", _("Integer"), ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("E1", _("Euros"), ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("F1", _("Dollars"), ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("G1", _("Percentage"), ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("H1", _("Number with 2 decimals"), ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("I1", _("Number with 6 decimals"), ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("J1", _("Time"), ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("K1", _("Boolean"), ColorsNamed.Orange, "BoldCenter")
    colors_list=([a for a in dir(ColorsNamed()) if not a.startswith('__')])
    for row, color_str in enumerate(colors_list):
        color_key=getattr(ColorsNamed(), color_str)
        doc.addCellWithStyle(C("A2").addRow(row), color_str, color_key, "Bold")
        doc.addCellWithStyle(C("B2").addRow(row), datetime.now(), color_key, "Datetime")
        doc.addCellWithStyle(C("C2").addRow(row), date.today(), color_key, "Date")
        doc.addCellWithStyle(C("D2").addRow(row), pow(-1, row)*-10000000, color_key, "Integer")
        doc.addCellWithStyle(C("E2").addRow(row), Currency(pow(-1, row)*12.56, "EUR"), color_key, "EUR")
        doc.addCellWithStyle(C("F2").addRow(row), Currency(pow(-1, row)*12345.56, "USD"), color_key, "USD")
        doc.addCellWithStyle(C("G2").addRow(row), Percentage(pow(-1, row)*1, 3), color_key,  "Percentage")
        doc.addCellWithStyle(C("H2").addRow(row), pow(-1, row)*123456789.121212, color_key, "Float6")
        doc.addCellWithStyle(C("I2").addRow(row), pow(-1, row)*-12.121212, color_key, "Float2")
        doc.addCellWithStyle(C("J2").addRow(row), (datetime.now()+timedelta(seconds=3600*12*row)).time(), color_key, "Time")
        doc.addCellWithStyle(C("K2").addRow(row), bool(row%2), color_key, "Bool")

    doc.addCellWithStyle(C("E2").addRow(row+1),f"=sum(E2:{C('E2').addRow(row).string()})", ColorsNamed.GrayLight, "EUR" )
    doc.addCellMergedWithStyle("E15:K15", "Prueba de merge", ColorsNamed.Yellow, style="BoldCenter")
    doc.setComment("B14", "This is nice comment")
    
    doc.freezeAndSelect("B2")
    
#    # Un comment to see objjects from cell
#    from unogenerator.commons import Coord_from_index
#    for i in range(10):
#        o=doc.getValue(Coord_from_index(i, 3), True)
#        print(o)

    ## HELPERS
    doc.createSheet("Helpers")
    doc.addCellMergedWithStyle("A1:E1","Helper values with total (horizontal)", ColorsNamed.Orange, "BoldCenter")
    helper_title_values_total_row(doc, "A2", "Suma 3", [1,2,3])

    doc.addCellMergedWithStyle("A4:E4","Helper values with total (vertical)", ColorsNamed.Orange, "BoldCenter")
    helper_title_values_total_column(doc, "A5", "Suma 3", [1,2,3])
    
    doc.addCellMergedWithStyle("A11:C11","List of rows", ColorsNamed.Orange, "BoldCenter")
    doc.addListOfRowsWithStyle("A12", [[1,2,3],[4,5,6],[7,8,9]], ColorsNamed.White)
    
    doc.addCellMergedWithStyle("E11:G11","List of columns", ColorsNamed.Orange, "BoldCenter")
    doc.addListOfColumnsWithStyle("E12", [[1,2,3],[4,5,6],[7,8,9]], ColorsNamed.White)

    helper_totals_row(doc, "A17", ["#SUM"]*3, styles=None, row_from="12", row_to="15")
    helper_totals_column(doc, "I12", ["#SUM"]*3, styles=None, column_from="E", column_to="G")


    doc.addCellMergedWithStyle("A19:D19","List of rows with totals", ColorsNamed.Orange, "BoldCenter")
    doc.addListOfRowsWithStyle("A20", [["A",12000,2,3],["B",1020,5,6],["C",20404,8,9]], ColorsNamed.White)
    helper_totals_from_range(doc, "B20:D22", totalcolumns=True, totalrows=True)

    
    doc.addCellMergedWithStyle("A25:B25","List of ordered dictionaries", ColorsNamed.Orange, "BoldCenter")
    lod=[]
    lod.append(OrderedDict({"Singer": "Elvis",  "Song": "Fever" }))
    lod.append(OrderedDict({"Singer": "Roy Orbison",  "Song": "Blue angel" }))
    helper_list_of_ordereddicts(doc, "A26",  lod, columns_header=1)
    
    doc.addCellMergedWithStyle("A30:B30","List of dictionaries", ColorsNamed.Orange, "BoldCenter")
    helper_list_of_dicts(doc, "A31",  lod, keys=["Song",  "Singer"])
    
    doc.addCellMergedWithStyle("A35:D35","List of ordered dictionaries one method with totals", ColorsNamed.Orange, "BoldCenter")
    lod=[]
    lod.append(OrderedDict({"Singer": "Elvis",  "Songs": 10000 , "Albums": 100}))
    lod.append(OrderedDict({"Singer": "Roy Orbison",  "Songs": 100,  "Albums": 20 }))
    helper_list_of_ordereddicts_with_totals(doc, "A36",  lod, columns_header=1)
    
    
    doc.addColumnWithStyle("A41",  [1, 3, 2, 4, 1, 5])
    doc.setColorScale("A41:A46")


#    # Un comment to see objjects from cell
#    l=doc.getValuesByRange("E19:E23", True)
#    for o in l:
#        print(o)


    ##Sort
    doc.createSheet("Sort")
    l=[7, 3, 2, 5, 6, 0, 9, 4, 10]
    doc.addCellWithStyle("A1",  "Unsorted", ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("B1", "Sorted ASC", ColorsNamed.Orange, "BoldCenter")
    doc.addCellWithStyle("C1", "Sorted DESC", ColorsNamed.Orange, "BoldCenter")
    doc.addColumnWithStyle("A2", l)
    doc.addColumnWithStyle("B2", l)
    doc.addColumnWithStyle("C2", l)
    doc.sortRange("B2:B10",  0)
    doc.sortRange("C2:C10",  0, False)
    
    ## Sheet with all styles names
    helper_ods_sheet_stylenames(doc)
    doc.statistics.ods_sheet_statistics()

    doc.save(f"unogenerator_example_{language}{suffix}.ods")
    doc.export_xlsx(f"unogenerator_example_{language}{suffix}.xlsx")
    doc.export_pdf(f"unogenerator_example_{language}{suffix}.pdf")
    doc.close()
    
    r= _(f"unogenerator_example_{language}{suffix}.ods took {datetime.now()-doc.statistics.init} in {port}")
    info(r)
    return r
    
    
def demo_odt_standard(language, port=2002, suffix=""):
    lang1=translation('unogenerator', resource_filename("unogenerator","locale"), languages=[language])
    lang1.install()
    _=lang1.gettext

    doc=ODT_Standard(port)
    doc.setMetadata(
        _("UnoGenerator documentation"),  
        _("Unogenerator python module documentation"), 
        "Turulomio", 
        _(f"This file have been generated with UnoGenerator-{__version__}. You can see UnoGenerator main page in http://github.com/turulomio/unogenerator"), 
        ["unogenerator", "demo", "files"]
    )
    
    
    doc.addParagraph(_("UnoGenerator documentation"), "Title")
    doc.addParagraph(_(f"Version: {__version__}"), "Subtitle")
    
    doc.addParagraph(_("Introduction"),  "Heading 1")
    
    doc.addParagraph(
        _("UnoGenerator uses Libreoffice UNO API python bindings to generate documents.") +" " +
        _("So in order to use, you need to launch a --headless libreoffice instance.") + " "+
        _("We recomend to launch then easily with 'unogenerator_start' script."), 
        "Standard"
    )

    doc.addParagraph(
        _("UnoGenerator has standard templates to help you with edition, although you can use your own templates.") +" " + 
        _("You can edit this one or create your own.")  +" " +
        _("This document has been created with 'standard.odt' files that you can find inside this python module."), 
        "Standard"
    )
        
    doc.addParagraph(_("Installation"), "Heading 2")
    doc.addParagraph(_("You can use pip to install this python package:") ,  "Standard")
    doc.addParagraph("""pip install unogenerator"""    , "Code")
    doc.addParagraph(_("ODT 'Hello World' example"), "Heading 2")
    doc.addParagraph(_("This is a Hello World example. You get the example in odt, docx and pdf formats:") ,  "Standard")
    doc.addParagraph("""from unogenerator import ODT_Standard
doc=ODT_Standard()
doc.addParagraph("Hello World", "Heading 1")
doc.addParagraph("Easy, isn't it","Standard")
doc.save("hello_world.odt")
doc.export_docx("hello_world.docx")
doc.export_pdf("hello_world.pdf")
doc.close()"""    , "Code")
    doc.pageBreak()
    
    doc.addParagraph(_("ODT"), "Heading 1")
    doc.addParagraph(
        _("ODT files can be quickly generated with UnoGenerator.") + " " + 
        _("There is a predefined template in code called 'standard.odt' to help you with edition."),  
        "Standard"
    )
    
    
    doc.addParagraph(_("Calling the ODT constructor"), "Heading 2")
    doc.addParagraph(_("You can call ODT constructor in this ways:") , "Standard")
        
    doc.addParagraph(
        _("ODT with standard template (Recomended).") + " " + 
        _("There is a predefined template in code called 'standard.odt', inside this python module, to help you with edition, although you can use your own ones.") +" "+
        _("With this mode you can create new documents"), 
        "BulletsLevel1"
    )
    
    doc.addParagraph("""from unogenerator import ODT_Standard
doc=ODT_Standard()"""    , "Code")

    doc.addParagraph(
        _("ODT with template or file (Recomended).") + " " + 
        _("With this mode you can read your files to overwrite them or use your file as a new template to create new documents"), 
        "BulletsLevel1"
    )
    
    doc.addParagraph("""from unogenerator import ODT
doc=ODT('yourdocument.odt')"""    , "Code")
    
    doc.addParagraph(
        _("ODT without template.") + " " + 
        _("With this mode you can write your files with Libreoffice default styles.") +" " +
        _("If you want to create new ones, you should write them using Libreoffice API code"), 
        "BulletsLevel1"
    )
    
    
    doc.addParagraph("""from unogenerator import ODT
doc=ODT()"""    , "Code")
    
    doc.addParagraph(_("Styles"), "Heading 2")
    doc.addParagraph(
        _("To call default Libreoffice paragraph styles you must use their english name.") + " " + 
        _("You can see their names with this method:"), "Standard"
    )
    doc.addParagraph("""doc.print_styles()"""    , "Code")

    doc.addParagraph(_("Tables"), "Heading 2")
    table_data=[
        [_("Concept"), _("Value"), _("Comment") ], 
        [_("Text"), _("This is a text"), _("Good")], 
        [_("Datetime"), datetime.now(), _("Good")], 
        [_("Date"), date.today(), _("Good")], 
        [_("Float"),  12.121, _("Good")], 
        [_("Currency"), Currency(-12.12, "EUR"), _("Good")], 
        [_("Percentage"), Percentage(1, 3), _("Good")], 
    ]
    
    columnspercentages=[15, 70, 15 ]
    doc.addParagraph(_("We can create tables with diferent font sizes and formats:") + str(columnspercentages), "Standard")
    doc.addTableParagraph(table_data, columnssize_percentages=columnspercentages, style="Table1")
    
    doc.addTableParagraph(table_data, columnssize_percentages=[30, 40,30],  size=6, style="Table1")
    

    doc.addParagraph(_("Lists and numbered lists"), "Heading 2") 
    doc.addParagraph(_("Simple list"), "BulletsLevel1")
    doc.addParagraph(_("Simple list"), "BulletsLevel2")
    doc.addParagraph(_("Simple list"), "BulletsLevel2")
    doc.addParagraph(_("Simple list"), "BulletsLevel1")
    doc.addParagraph(_("Simple list"), "BulletsLevel2")
    doc.addParagraph(_("Simple list"), "BulletsLevel1")


    doc.pageBreak()
    doc.addParagraph(_("Images"), "Heading 2")
    
    l=[]
    l.append( _("Este es un ejemplo de imagen as char: "))
    l.append(doc.textcontentImage(resource_filename(__name__, 'images/crown.png'), 1, 1, "AS_CHARACTER", "PRIMERA", linked=False))
    l.append(". Ahora sigo escribiendo sin problemas.")
    doc.addParagraphComplex(l, "Standard")
    
    l=[]
    l.append( _("This is an image loaded from bytes: "))
    l.append(doc.textcontentImage(open(resource_filename(__name__, 'images/crown.png'), "rb").read(), 1, 1, "AS_CHARACTER", "d", linked=False))
    doc.addParagraphComplex(l, "Standard")

    l=[]
    l.append( _("As you can see, I can reuse it one hundred times. File size will not be increased because I used reference names."))
    for i in range(100):
        l.append(doc.textcontentImage(resource_filename(__name__, 'images/crown.png'), 0.5,  0.5, "AS_CHARACTER", linked=False))
    doc.addParagraphComplex(l, "Standard")


    doc.addParagraph(_("The next paragraph is generated with the illustration method"), "Standard")
    doc.addImageParagraph([resource_filename(__name__, 'images/crown.png')]*5, 2.5, 1.5, "Illustration", linked=False)


    doc.addParagraph(_("Search and Replace"), "Heading 2")
    doc.addParagraph(_("Below this paragraph is a paragraph with a % REPLACEME % (Without white spaces) text and it's going to be replaced after all document is been generated"), "Standard")
    doc.addParagraph("%REPLACEME%", "Standard")

    doc.pageBreak()
    doc.addParagraph(_("ODS"), "Heading 1")

    doc.addParagraph(_("ODS 'Hello World' example"), "Heading 2")
    doc.addParagraph(_("This is a Hello World example. You'll get the example in ods, xlsx and pdf formats:") ,  "Standard")
    doc.addParagraph("""from unogenerator import ODS_Standard
doc=ODS_Standard()
doc.addCellMergedWithStyle("A1:E1", "Hello world", style="BoldCenter")
doc.save("hello_world.ods")
doc.export_xlsx("hello_world.xlsx")
doc.export_pdf("hello_world.pdf")
doc.close()"""    , "Code")
    
    doc.find_and_replace("%REPLACEME%", _("This paragraph was set at the end of the code after a find and replace command."))
    doc.paragraphBreak()
    doc.addParagraph(_("This paragraph was set after replacement."), "Standard")
    doc.pageBreak()
    doc.addParagraph(_("This paragraph was set after a page break."), "Standard")
    doc.pageBreak("Landscape")
    doc.addParagraph(_("This paragraph was set after a page break with Landscape style."), "Standard")
    
    
#    doc.find_and_delete_until_the_end_of_document("This paragraph was set after replacement.")

    doc.addParagraph(_("This is a pair of brackets ()."), "Standard")
    doc.findall_and_replace(" ().", " )(.", True)
    
    doc.addParagraph(_("NOW)("), "Standard")
    doc.addParagraph(_("This is a set of symbols: .,:;?ºª\_-/()."), "Standard")
    doc.findall_and_replace(".,:;?ºª\_-/().", ".,:;?ºª\_-/(). REPLACED", True)
    doc.addParagraph(_("NOW)("), "Standard")
    
    
    
    doc.save(f"unogenerator_documentation_{language}{suffix}.odt")
    doc.export_docx(f"unogenerator_documentation_{language}{suffix}.docx")
    doc.export_pdf(f"unogenerator_documentation_{language}{suffix}.pdf")
    doc.close()
    r=_(f"unogenerator_documentation_{language}{suffix}.odt took {datetime.now()-doc.statistics.init} in {port}")
    info(r)
    return r

if __name__ == "__main__":
    main()



from datetime import datetime, timedelta, date
from decimal import Decimal
from unogenerator import ODS_Standard
from unogenerator.commons import Coord, ColorsNamed
from unogenerator.reusing.percentage import Percentage
from unogenerator.reusing.currency import Currency


number=2000
lor=[]
for i in range(number):
    lor.append([
        i, 
        Coord("B2").addRow(i).string(), 
        date.today()+timedelta(days=i) , 
        datetime.now()+timedelta(hours=i), 
        1.123*i, 
        Decimal(1.123*i), 
        Percentage(1, 1.123*i), 
        Currency(1.123*i, "EUR"), 
        bool(i % 2), 
    ])
headers=["N", "String", "Date", "Datetime", "Float", "Decimal", "Percentage", "Currency", "Boolean"]


start=datetime.now()
doc=ODS_Standard()
doc.createSheet("Benchmark")
doc.setColumnsWidth([4]*20)
doc.addRowWithStyle("A1", headers, ColorsNamed.Orange, "BoldCenter")
doc.addListOfRowsWithStyle(Coord("A2"), lor, cellbycell=True)
print(f"Unogenerator creates {number} cells (cellbycell) in {datetime.now()-start}")
doc.save("benchmark_unogenerator_guessing_styles.ods")
doc.close()

start=datetime.now()
doc=ODS_Standard()
doc.createSheet("Benchmark")
doc.setColumnsWidth([4]*20)
doc.addRowWithStyle("A1", headers, ColorsNamed.Orange, "BoldCenter")

doc.addListOfRowsWithStyle(Coord("A2"), lor)
print(f"Unogenerator creates {number} cells in {datetime.now()-start}")
doc.save("benchmark_unogenerator_using_massive.ods")
doc.close()



from argparse import ArgumentParser,  RawTextHelpFormatter
from gettext import translation
from os import path,  makedirs
from polib import POEntry,  POFile, pofile
from subprocess import run
from pkg_resources import resource_filename
from unogenerator.commons import __version__, argparse_epilog
from unogenerator import ODT


try:
    t=translation('unogenerator', resource_filename("unogenerator","locale"))
    _=t.gettext
except:
    _=str

def run_check(command, shell=False):
    p=run(command, shell=shell, capture_output=True);
    if p.returncode!=0:
        print(f"Error en comando. {command}")
        print("STDOUT:")
        print(p.stdout.decode('utf-8'))
        print("STDERR:")
        print(p.stderr.decode('utf-8'))
        print("Saliendo de la instalación")
        exit(2)

            
def main():
    parser=ArgumentParser(prog='unogenerator', description=_('Generate a po and pot file from ODF'), epilog=argparse_epilog(), formatter_class=RawTextHelpFormatter)
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--from_language', action='store', help=_('Language to translate from. Example codes: es, fr, en, md'), required=True, metavar="CODE")
    parser.add_argument('--to_language', action='store', help=_('Language to translate to. Example codes: es, fr, en, md'), required=True,  metavar="CODE")
    parser.add_argument('--input', action='append', help=_('Files to translate. You can set several files.'), required=True,  metavar="FILE")
    parser.add_argument('--output_directory', action='store', help=_('Output directory with results and catalogues'), required=True,  metavar="FILE")
    parser.add_argument('--undetected', action='append', help=_('Undetected strings to append to translation'), default=[])
    parser.add_argument('--translate', action='store_true', help=_('Creates translated file'), default=False)
    parser.add_argument('--fake', action='store_true', help=_('Sets a fake translation to all strings'), default=False)
    parser.add_argument('--pdf', action='store_true', help=_('Creates translated file in pdf'), default=False)
    args=parser.parse_args()
    
    command(args.from_language, args.to_language, args.input, args.output_directory, args.translate, args.undetected, args.fake, args.pdf)

def same_entries_to_ocurrences(l):
    l= sorted(l, key=lambda x: (x[0], x[1], x[2], x[3]))
    r=[]
    for filename, type, number,  position,  text in l:
        r.append((filename, f"{type}#{number}#{position}"))
    return r

    
def getEntriesFromDocument(filename):
        r=[]
        doc=ODT(filename)

        #Extract strings from paragraphs
        r=r+entries_from_paragraph_enumeration("Paragraph", doc.cursor.Text.createEnumeration(), filename)

        for style in doc.getPageStylesObjects():            
            #Extract strings from headers
            ht=style.HeaderText
            if ht is None:
                continue
            r=r+entries_from_paragraph_enumeration("HeaderParagraph", ht.createEnumeration(), filename)
            #Extract strings from foot
            ft=style.FooterText
            if ft is None:
                continue
            r=r+entries_from_paragraph_enumeration("FooterParagraph", ft.createEnumeration(), filename)
        doc.close()
        return r
        
def entries_from_paragraph_enumeration(title, enumeration, filename):
        r=[]
        for i,  par in enumerate(enumeration):
#            print(i, par, dir(par))
            if  par.supportsService("com.sun.star.text.Paragraph") :
                for position, element in enumerate(par.createEnumeration()):
                    text_=element.getString()
                    if text_ !="" and text_!=" " and text_!="  ":
                        entry=(filename, title,  i,  position, text_)
#                        print(entry)
                        r.append(entry)
            elif  par.supportsService("com.sun.star.text.TextTable") :
#                print("AQUQI", par, dir(par), par.getColumns().Count, par.getRows().Count)
                for column in range(par.getColumns().Count):
                    for row in range(par.getRows().Count):
                        try:
                            cell=par.getCellByPosition(column, row)
                            r=r+entries_from_paragraph_enumeration(f"{title}Table,paragraph{i},column{column},row{row}", cell.createEnumeration(), filename)
                        except:
                            print (_("Perhaps a merged string"))
        return r
            


def generate_pot_file(potfilename, set_strings, entries):
    file_pot = POFile()
    file_pot.metadata = {
        'Project-Id-Version': '1.0',
        'Report-Msgid-Bugs-To': 'you@example.com',
        'POT-Creation-Date': '2007-10-18 14:00+0100',
        'PO-Revision-Date': '2007-10-18 14:00+0100',
        'Last-Translator': 'you <you@example.com>',
        'Language-Team': 'English <yourteam@example.com>',
        'MIME-Version': '1.0',
        'Content-Type': 'text/plain; charset=utf-8',
        'Content-Transfer-Encoding': '8bit',
    }
    for s in set_strings:
        same_entries=[] #Join seame text entries
        for filename, type, number, position, string_ in entries:
            if string_==s:
                same_entries.append((filename, type, number, position, string_))

        entry = POEntry(
            msgid=s,
            msgstr='', 
            occurrences=same_entries_to_ocurrences(same_entries)
        )
        file_pot.append(entry)
    file_pot.save(potfilename)

def command(from_language, to_language, input, output_directory, translate,  undetected_strings=[], fake=False, pdf=False):   
    makedirs(output_directory, exist_ok=True)
    makedirs(f"{output_directory}/{to_language}", exist_ok=True)
    
    pot=f"{output_directory}/catalogue.pot"
    po=f"{output_directory}/{to_language}/{to_language}.po"
        
    entries=[]#List of (filename,"type", numero, posicion) type=Paragraph, numero=numero parrafo y posición orden dentro del parrafo
    print(_("Extracting strings from:"))
    for filename in input:
        print(_(f"   - {filename}"))
        entries=entries+getEntriesFromDocument(filename)
        
    # Set with distinct strings of all entries
    set_strings=set()
    for filename, type, number, position, string_ in entries:
        set_strings.add(string_)

    #Generate pot file
    generate_pot_file(pot, set_strings, entries)
    
    #Merging pot with out po file
    if path.exists(po)==False:
        run_check(["msginit", "-i", pot,  "-o", po])
    run_check(["msgmerge","-N", "--no-wrap","-U", po, pot])
    
    print(f"{len(set_strings)} different strings detected")
    
        
    # Creates a dictionary of translations
    dict_po={}
    for i, entry in enumerate(pofile(po)):
        if fake is True:
            dict_po[entry.msgid]=f"{{{entry.msgid}}}"
        else:
            if entry.msgstr == "":
                dict_po[entry.msgid]=entry.msgid
            else:
                dict_po[entry.msgid]=entry.msgstr
    
    if translate is True:
        print(_("Translating files to:"))
        for filename_input in input:
            output=f"{output_directory}/{to_language}/{path.basename(filename_input)}"
            print(_(f"   + {output}"))
            write_translation(filename_input, output,  dict_po, entries, pdf)
            
            
def write_translation(original, filename, dict_po, entries, pdf):
    doc=ODT(original)
    search_descriptor=None
    replaced=0
    for filename_, type, number,  position,  text in entries:
        if filename_==path.basename(filename):
            search_descriptor=doc.find_and_replace(text, dict_po[text], search_descriptor)
#            print(f"{text} ==> {dict_po[text]}")
            if search_descriptor is not None:
                replaced=replaced+1
    print(f"      - Translated {replaced} strings")
    doc.save(filename)
    if pdf is True:
        print("      - PDF translation generated")
        doc.export_pdf(filename[:-4] +".pdf")
    doc.close()
        




from os import path
from uno import getComponentContext, createUnoStruct, systemPathToFileUrl
from com.sun.star.beans import PropertyValue
localContext = getComponentContext()

resolver = localContext.ServiceManager.createInstance('com.sun.star.bridge.UnoUrlResolver')
ctx = resolver.resolve('uno:socket,host=127.0.0.1,port=2002;urp;StarOffice.ComponentContext')
desktop = ctx.ServiceManager.createInstance('com.sun.star.frame.Desktop')


args=(
    PropertyValue('AsTemplate',0,True,0),
)
template=systemPathToFileUrl(path.abspath("colorscale.ods"))
document=desktop.loadComponentFromURL(template,'_blank', 8, args)
css=ctx.ServiceManager.createInstance("com.sun.star.sheet.ColorScale")    
print("css", css, dir(css))

print("document",document,dir(document))

sheet=document.getSheets().getByIndex(0)

print("Sheet",sheet,dir(sheet))

# I create and select a range of cells
oRangos = document.createInstance("com.sun.star.sheet.SheetCellRanges")
oRangos.addRangeAddress( sheet.getCellRangeByName( "A1:A8" ).getRangeAddress() ,False )
document.getCurrentController().select(oRangos)

print("oRangos",oRangos,dir(oRangos))

cf=oRangos.ConditionalFormat

print(cf.Count)

print("RangoCF", cf, dir(cf))
        
# #I create a conditional format range in the sheet
# cfs=sheet.ConditionalFormats.createByRange(oRangos)
# cf=sheet.ConditionalFormats.getConditionalFormats()[cfs-1]

# # I Create a colorscale conditional format

# cf.createEntry(1, 0)   #1 iS COLORSCALE
# print(cf,dir(cf))
# entry=cf[0]
# print("entry", entry,dir(entry),entry.getPropertySetInfo())
# # # pep=entry.queryInterface("com.sun.star.sheet.XColorScaleEntry")
# # # print(pep,dir(pep))
# cse=entry.ColorScaleEntries
# print("CSENTRIES",cse,dir(cse),cse.__class__)


# from com.sun.star.beans import PropertyValue

# from com.sun.star.sheet import XColorScaleEntry
# celladdress= createUnoStruct("com.sun.star.table.CellAddress")
# celladdress= createUnoStruct("com.sun.star.sheet.XColorScaleEntry")
# print("CELLADDRESS", celladdress,dir(celladdress))
# a=XColorScaleEntry()
# print("A",a,dir(a))
# a.setType(1)


# colorentry=list()

# entry= list()
# entry.append(PropertyValue("Formula",0,0,0))
# entry.append(PropertyValue("Color",0,16711680,0))
# entry.append(PropertyValue("Type",0,0,0))
# colorentry.append(tuple(entry))
# entry= list()
# entry.append(PropertyValue("Formula",0,50,0))
# entry.append(PropertyValue("Color",0,16777215,0))
# entry.append(PropertyValue("Type",0,2,0))
# colorentry.append(tuple(entry))
# entry= list()
# entry.append(PropertyValue("Formula",0,0,0))
# entry.append(PropertyValue("Color",0,43315,0))
# entry.append(PropertyValue("Type",0,1,0))
# colorentry.append(tuple(entry))

# colorentry=tuple(colorentry)
# cse=colorentry

# for o in cse:
#     print("cse",o, dir(o))


# I don't know how to follow
#new = createUnoStruct()


save_args=(
    PropertyValue('FilterName', 0, 'calc8', 0),
)
document.storeAsURL(systemPathToFileUrl(path.abspath("output.ods")), (save_args))
document.dispose()



from os import path
from uno import getComponentContext, createUnoStruct, systemPathToFileUrl
from com.sun.star.beans import PropertyValue
from com.sun.star.sheet.ConditionEntryType import COLORSCALE
localContext = getComponentContext()

resolver = localContext.ServiceManager.createInstance('com.sun.star.bridge.UnoUrlResolver')
ctx = resolver.resolve('uno:socket,host=127.0.0.1,port=2002;urp;StarOffice.ComponentContext')
desktop = ctx.ServiceManager.createInstance('com.sun.star.frame.Desktop')


args=(
    PropertyValue('AsTemplate',0,True,0),
)
template=systemPathToFileUrl(path.abspath("colorscale.ods"))
document=desktop.loadComponentFromURL(template,'_blank', 8, args)

sheet=document.getSheets().getByIndex(0)

# I create and select a range of cells
oRangos = document.createInstance("com.sun.star.sheet.SheetCellRanges")
oRangos.addRangeAddress( sheet.getCellRangeByName( "B1:B8" ).getRangeAddress() ,False )
document.getCurrentController().select(oRangos)
        
#I create a conditional format range in the sheet


cfs=sheet.ConditionalFormats
print("CFS", cfs, dir(cfs),"\n")
cf_id=sheet.ConditionalFormats.createByRange(oRangos)
print(cf_id,"\n")
cf=sheet.ConditionalFormats.getConditionalFormats()[cf_id-1]

print("CF", cf, dir(cf),cf.Count,"\n")

# I Create a colorscale conditional format

cf.createEntry(COLORSCALE, 0)   
print("CFDESPUES CREATEENTRY",cf,dir(cf), cf.Count,"\n")
entry=cf[0]

print("entry", entry,dir(entry),entry.getPropertySetInfo(),"\n")


for p in entry.getPropertySetInfo().Properties:
    print("P" , p, dir(p) , p.value, "\n")



# # pep=entry.queryInterface("com.sun.star.sheet.XColorScaleEntry")
# # print(pep,dir(pep))
# cse=entry.ColorScaleEntries
# print("CSENTRIES",cse,dir(cse),cse.__class__,"\n")


# from com.sun.star.beans import PropertyValue

from com.sun.star.sheet import XColorScaleEntry
# celladdress= createUnoStruct("com.sun.star.table.CellAddress")
# celladdress= createUnoStruct("com.sun.star.sheet.XColorScaleEntry")
# print("CELLADDRESS", celladdress,dir(celladdress))
a=XColorScaleEntry()
print("A",a,dir(a))
# a.setType(1)


colorent=list()

ent= list()
ent.append(PropertyValue("Formula",0,0,0))
ent.append(PropertyValue("Color",0,16711680,0))
ent.append(PropertyValue("Type",0,0,0))
colorent.append(tuple(ent))
ent= list()
ent.append(PropertyValue("Formula",0,50,0))
ent.append(PropertyValue("Color",0,16777215,0))
ent.append(PropertyValue("Type",0,2,0))
colorent.append(tuple(ent))
ent= list()
ent.append(PropertyValue("Formula",0,0,0))
ent.append(PropertyValue("Color",0,43315,0))
ent.append(PropertyValue("Type",0,1,0))
colorent.append(tuple(ent))

colorent=tuple(colorent)

entry.setPropertyValue("ColorScaleEntries", colorent)



# for o in cse:
#     print("cse",o, dir(o))


# I don't know how to follow
#new = createUnoStruct()


save_args=(
    PropertyValue('FilterName', 0, 'calc8', 0),
)
document.storeAsURL(systemPathToFileUrl(path.abspath("output.ods")), (save_args))
document.dispose()



## @brief my Percentage object
## THIS IS FILE IS FROM https://github.com/turulomio/reusingcode IF YOU NEED TO UPDATE IT PLEASE MAKE A PULL REQUEST IN THAT PROJECT
## DO NOT UPDATE IT IN YOUR CODE IT WILL BE REPLACED USING FUNCTION IN README

from logging import warning
from decimal import Decimal

## Class to manage percentages in spreadsheets
class Percentage:
    def __init__(self, numerator=None, denominator=None):
        self.setValue(self.toDecimal(numerator),self.toDecimal(denominator))




    def toDecimal(self, o):
        if o==None:
            return o
        if o.__class__.__name__ in ["Currency", "Money"]:
            return o.amount
        elif o.__class__.__name__=="Decimal":
            return o
        elif o.__class__.__name__ in ["int", "float"]:
            return Decimal(o)
        elif o.__class__.__name__=="Percentage":
            return o.value
        else:
            warning (o.__class__.__name__)
            return None

    def __repr__(self):
        return self.string()

    def __neg__(self):
        if self.value==None:
            return self
        return Percentage(-self.value, 1)

    def __lt__(self, other):
        if self.value==None:
            value1=Decimal('-Infinity')
        else:
            value1=self.value
        if other.value==None:
            value2=Decimal('-Infinity')
        else:
            value2=other.value
        if value1<value2:
            return True
        return False



    def __add__(self,p):
        return self.__class__(self.value+p.value,1)

    def __sub__(self, p):
        return self.__class__(self.value-p.value,1)

    def __mul__(self, value):
        if self.value==None or value==None:
            r=None
        else:
            r=self.value*self.toDecimal(value)
        return Percentage(r, 1)

    def __truediv__(self, value):
        try:
            r=self.value/self.toDecimal(value)
        except:
            r=None
        return Percentage(r, 1)

    def setValue(self, numerator,  denominator):
        try:
            self.value=Decimal(numerator/denominator)
        except:
            self.value=None

    def value_100(self):
        if self.value==None:
            return None
        else:
            return self.value*Decimal(100)

    ## @return percentage float value
    def float_100(self):
        return float(self.value_100())

    def string(self, rnd=2):
        if self.value==None:
            return "None %"
        return "{} %".format(round(self.value_100(), rnd))

    ## Returns if the percentage is valid. I mean it's value different of None
    def isValid(self):
        if self.value!=None:
            return True
        return False

    def isGETZero(self):
        if self.isValid() and self.value>=0:
            return True
        return False

    def isGTZero(self):
        if self.isValid() and self.value>0:
            return True
        return False

    def isLTZero(self):
        if self.isValid() and self.value<0:
            return True
        return False

    def qtablewidgetitem(self, decimals=2):
        from .. ui.myqtablewidget  import qpercentage
        return qpercentage(self, decimals=2)

## Calculates porcentage to pass from a to b
## @param a. Can be an object divisible and that can be substracted
def percentage_between(a,b):
    try:
        return Percentage(b-a,a)
    except:
        return Percentage()
